#include <stdio.h>			/* Standard I/I */
#include <cave_ogl.h>
#include <c2c.h>                        /* CAVE-to-CAVE library */

#define VERTEX_STREAM 10              /* Requested stream id */
#define BUTTON_STREAM 11              /* Requested stream id */

void UserData(void *data, CAVEId id); /* Callback prototype */

void simple_draw(void);
void init_shared_data();

#define BUFF_SIZE 500000
struct global_mem {
	int ix;
	float vertx[BUFF_SIZE];
	float verty[BUFF_SIZE];
	float vertz[BUFF_SIZE];
} *gm;

main(int argc,char **argv)
{

	int rc;
	c2cWorldDataInit();
	CAVEConfigure(&argc,argv,NULL);
	init_shared_data();
	CAVEInit();
	CAVEDisplay(simple_draw,0);
	c2cWorldInit(&argc,argv);              /* Start CAVE-to-CAVE functionality */
	c2cRegisterStream(BUTTON_STREAM,NULL,NULL);

	rc = -1;
	while( rc != E_SUCCESS ) {
		rc = c2cWorldDataSubscribe("Monte Carlo PI",VERTEX_STREAM,UserData);
	}


	 while (!CAVEgetbutton(CAVE_ESCKEY)) {
		if (CAVEButtonChange(1)  == 1 || CAVEButtonChange(3) == 1) {
			printf("%d, %d\n", CAVEBUTTON1, CAVEBUTTON3);
			if (CAVEBUTTON1)
				send_button(1);
			if (CAVEBUTTON3)
				send_button(3);
		}
		c2cUpdate();
	  }
	   c2cWorldExit();

	 CAVEExit();
}

send_button(x)
int x;
{
	c2cBuffer *databuffer;                /* Data buffer to pack */

	printf("Sending %d\n",x);

	c2cInitPackBuffer(&databuffer);     /* Initialize pack buffer */

	c2cPackInt(&databuffer,&x,1);
	c2cWorldSendStream(&databuffer,BUTTON_STREAM);
	c2cFreePackBuffer(&databuffer); 
}
float v1[3] = { -1.0,  4.0, -4.0 };
float v2[3] = {  1.0,  4.0, -4.0 };
float v3[3] = {  0.0,  6.0, -4.0 };


void simple_draw(void)
{
    GLint tri;

	int i;
	float vert1[3] = { -1.0,  4.0, -4.0 };
	float vert2[3] = {  1.0,  4.0, -4.0 };
	float vert3[3] = {  0.0,  6.0, -4.0 };
	float vert4[3] = {  0.0,  6.0, -4.0 };
	float side = 0.1;
  
    glClearColor(0.0, 0.0, 0.0, 0.0);
    glClearDepth(1.0);
    glClear(GL_COLOR_BUFFER_BIT | GL_DEPTH_BUFFER_BIT);
  
    glColor3f(1.0, 0.0, 0.0);
    
    for (i = 0; i< gm->ix; i++) {
	    glBegin(GL_POLYGON);

		  vert1[0] =  gm->vertx[i];
		  vert1[1] =  gm->verty[i] + 4;
		  vert1[2] =  gm->vertz[i];

		  vert2[0] = gm->vertx[i];
		  vert2[1] = gm->verty[i] + side + 4;
		  vert2[2] = gm->vertz[i];

		  vert3[0] = gm->vertx[i] + side;
		  vert3[1] = gm->verty[i] + side + 4;
		  vert3[2] = gm->vertz[i];

		  vert4[0] = gm->vertx[i] + side;
		  vert4[1] = gm->verty[i] + 4;
		  vert4[2] = gm->vertz[i];

		  glVertex3fv(vert1); 
		  glVertex3fv(vert2);
		  glVertex3fv(vert3);
		  glVertex3fv(vert4);

	    glEnd();
    }
}

/*
  This is the callback called when the stream data is received
*/
void UserData(void *databuffer, CAVEId id)
{ /* UserData */
  int type;
  double xyz[3];

  c2cGetDouble(databuffer,xyz,3);
  gm->vertx[gm->ix] =  xyz[0];
  gm->verty[gm->ix] =  xyz[1];
  gm->vertz[gm->ix] =  xyz[2];
  if (++gm->ix == BUFF_SIZE) {
	printf("Exceeded buff size \n");
	exit();
  }

  /*
  printf("                 double %f, %f, %f\n",xyz[0], xyz[1], xyz[2]);
  printf("from source %d\n",id);
  */
} /* UserData */

void
init_shared_data()
{


        gm = (struct global_mem *) CAVEMalloc(4096 + sizeof(struct global_mem));
        if (gm == NULL) {
                printf("Error in gm amalloc\n");
                exit(0);
        }

	gm->ix = 0;
  printf("After data init\n");
}


