/* compute pi using Monte Carlo method */

#define MPE_GRAPHICS

/* uncomment this to get two-d graphics 
 # define MPEGRAPHICS
*/

#include <stdio.h>
#include <math.h>
#include <sys/time.h>
#include "mpi.h"
#include "mpe.h"
#include "../simlib.h"

#define CHUNKSIZE     50*3
#define WINDOW_SIZE     500
#define RANDOM_SEED       0

#define VERTEX_STREAM 10              /* Requested stream id */
#define BUTTON_STREAM 11              /* Requested stream id */


/* message tags */
#define NEED_NUMBERS      1
#define RANDOM_NUMBERS    2

void button_callback();
int button = 0;

main(argc,argv)
int argc;
char *argv[];
{
    register int in,out,i,iters,max;
    register int ix,iy,iz;
    double x,y,z,calculatedPi;
    float  xyz[3];
    double error, epsilon;
    int ranks[1], done, temp;
    int gotEpsilon;
    int numprocs, myid, randServer, totalin, totalout, workerNum;
    MPI_Comm world, workers;
    MPI_Status mesgStatus;
    int randNums[CHUNKSIZE], request;
#ifdef MPEGRAPHICS    
    MPE_XGraph graph;
#endif
    int rc;
    SIM_Ctx *simctx;

    MPI_Init(&argc,&argv);
    world = MPI_COMM_WORLD;
    MPI_Comm_size(world,&numprocs);
    MPI_Comm_rank(world,&myid);
    randServer = numprocs-1;

    if (myid == 0) {
	printf("3 dim version\n");
	if ((argc!=2) || !sscanf( argv[1], "%lf", &epsilon )) {
	    printf( "Syntax:\n    %s <epsilon>\n", argv[0] );
	    gotEpsilon = 0;
	}
	else {
	    gotEpsilon = 1;
	}
    }
    MPI_Bcast( &gotEpsilon, 1, MPI_INT, 0, MPI_COMM_WORLD );
    if (!gotEpsilon) {
	MPI_Finalize();
	exit(1);
    }

#ifdef MPEGRAPHICS
    MPE_Open_graphics( &graph, MPI_COMM_WORLD, 0, -1, -1, WINDOW_SIZE,
		       WINDOW_SIZE, 0 );
#endif
    MPI_Comm_split( world, myid == randServer, 0, &workers );

    if (myid == randServer)	/* I am the rand server */
    {
#if RANDOM_SEED
        struct timeval time;
        gettimeofday( &time, 0 );
	/* initialize the random number generator */
	srandom( (int)(time.tv_usec*1000000+time.tv_sec) );
#endif
	MPI_Recv(&request, 1, MPI_INT, MPI_ANY_SOURCE, NEED_NUMBERS,
		 world, &mesgStatus);
	while ( request > 0 ) {
	    for (i = 0; i < CHUNKSIZE; i++)
		randNums[i] = random();
	    MPI_Send(randNums, CHUNKSIZE, MPI_INT, mesgStatus.MPI_SOURCE, 
		     RANDOM_NUMBERS, world);
	    MPI_Recv(&request, 1, MPI_INT, MPI_ANY_SOURCE, NEED_NUMBERS,
		     world, &mesgStatus);
	}
    }
    else			/* I am a worker process */
    {
        MPI_Comm_rank( workers, &workerNum );

	/* Contact the graphics server */
	simctx = SIM_Contact( "monte", "xfip" );
	if (!simctx) {
	    printf( "Could not contact xfip\n" );
	    MPI_Abort( MPI_COMM_WORLD, 1 );
	}
	/* Setup read callback for controls */
	if (workerNum == 0)
	    SIM_ReadCallback( simctx, button_callback );
	printf("Connected\n");

        request = 1;
	done = in = out = 0;
	max = ((unsigned) 1 << 31) - 1;	/*find max integer for normalization */
        MPI_Send( &request, 1, MPI_INT, randServer, NEED_NUMBERS, world );
	  /* request a string of random numbers */
	while (!done)
	{
	    request = 1;
	    MPI_Recv( randNums, CHUNKSIZE, MPI_INT, randServer, RANDOM_NUMBERS,
		     world, &mesgStatus );

	    for (i=0; i<CHUNKSIZE; )
	    {
	        x = (((double) randNums[i++])/max) * 2 - 1;
		y = (((double) randNums[i++])/max) * 2 - 1;
		z = (((double) randNums[i++])/max) * 2 - 1;
		if (x*x + y*y + z*z < 1.0)
		{
		    in++;
#ifdef MPEGRAPHICS
		    MPE_Draw_point( graph,
				  (int)(WINDOW_SIZE/2 + x*WINDOW_SIZE/2), 
				  (int)(WINDOW_SIZE/2 - y*WINDOW_SIZE/2),
				  MPE_BLACK );
#endif
		    xyz[0] = x;
		    xyz[1] = y;
		    xyz[2] = z;
		    SIM_SendFloats( simctx, xyz, 3 );
		}
		else
		    out++;
	    }
	    SIM_Update( simctx );
#ifdef MPEGRAPHICS
	    MPE_Update( graph );
#endif
	    temp = in;
	    MPI_Reduce(&temp, &totalin, 1, MPI_INT, MPI_SUM, 0, workers);
	      /* count total of ins  */ 
	    temp = out;
	    MPI_Reduce(&temp, &totalout, 1, MPI_INT, MPI_SUM, 0, workers);
	      /* count total of outs */

	    if (myid == 0)
	    {
	        calculatedPi = (6.0*totalin)/(totalin + totalout);
		error = fabs( calculatedPi-3.141592653589793238462643);
		done = (error < epsilon) || (totalin+totalout) > 1000000;
		printf( "pi = %23.20lf\n", calculatedPi );
		request = (done) ? 0 : 1;
		MPI_Send( &request, 1, MPI_INT, randServer, NEED_NUMBERS,
			  world );
	        MPI_Bcast( &done, 1, MPI_INT, 0, workers );
	    }
	    else
	    {
	        MPI_Bcast( &done, 1, MPI_INT, 0, workers );
		if (!done)
		{
		    request = 1;
		    MPI_Send( &request, 1, MPI_INT, randServer, NEED_NUMBERS,
			     world );
		}
	    }
	    if (!done && button == 3) {
		done = 1;
		/* Need to tell everyone else.  Do this by telling rand
		 server */
		request = 0;
		MPI_Send( &request, 1, MPI_INT, randServer, NEED_NUMBERS, 
			  world );
	    }
	}
	SIM_Close( simctx );
    }

    if (myid == 0)
    {
        printf( "\npoints: %d\nin: %d, out: %d\n", totalin+totalout,
	       totalin, totalout );
	printf( "Press almost any key to exit\n");
	getchar();
    }
#ifdef MPEGRAPHICS
    MPE_Close_graphics( &graph );
#endif
    MPI_Finalize();
}

void button_callback( simctx )
SIM_Ctx *simctx;
{
    SIM_ReadInts( simctx, &button, 1 );
    printf("Got %d\n",button);
}
	
