#include "pexec.h"
#include <stdio.h>
#include <errno.h>
#include "srlist.h"

SrList *Names = 0;

#define STDIN 0
#define STDOUT 1
/*
 * This program in a "finder" demon.  It takes key/name pairs and
 * provides install, lookup, delete, and list services.
 * Installed names have timeouts so that "old" data can be deleted.
 *
 * A simple console is also available (see scommand.c)
 *
 * Note that ALL interchanges with this demon use the command-line
 * interface.  Not only is this simpler (one interface), but it
 * ensures that all clients are on the same footing (any service added
 * to one is immediately part of the other).
 */

int GetConnection ANSI_ARGS((PX_Ctx *, PX_Conn *, PX_Handler, int )), 
    UnlinkLocal ANSI_ARGS((PX_Ctx *, PX_Conn *, PX_Handler, int ));
extern int ReadCommand ANSI_ARGS(( PX_Ctx *, PX_Conn *, PX_Handler, int));
extern void *InitReadCommand ANSI_ARGS(( PX_Ctx *, int ));

int main( argc, argv )
int  argc;
char **argv;
{
    PX_Ctx     ctx;
    PX_Handler close_fd, err_fd;
    PX_Handler read_command;
    PX_Handler unlinklocal, getconnection;
    int        fd;
    void       *sock_info;
    int        netsocket = FINDERPORT;

    PX_Ctx_init( &ctx );
    PX_Monitor_setup( &ctx );

    /* Create name table */
    Names = NewSrList( 512 );

    close_fd      = PX_Handler_simple_create( PX_Hfcn_close_fd, (void *)0 );
    err_fd        = PX_Handler_simple_create( PX_Hfcn_error_fd, (void *)0 );

/* Add a command interpreter to stdin */
    read_command = PX_Handler_simple_create( ReadCommand, 
					     InitReadCommand(&ctx, STDOUT ) );
    PX_FD_add( STDIN,  PX_READ, read_command, err_fd, close_fd, 
	       (struct timeval *)0, &ctx );

/* Add a local socket for a console */
    fd = PX_FD_Create_local_socket( LOCALPORT, 1, &sock_info );
    if (fd < 0) {
	fprintf( stderr, "Error creating local socket\n" );
    }
    else {
	getconnection = PX_Handler_create( GetConnection, sock_info, 
					   (int (*)())0,
					   (void *)0, UnlinkLocal );
	unlinklocal   = PX_Handler_simple_create( UnlinkLocal, (void *)0 );
	PX_FD_add( fd, PX_READ, getconnection, unlinklocal, unlinklocal, 
		   (struct timeval *)0, &ctx );
	PX_Handler_free( unlinklocal );
	PX_Handler_free( getconnection );
    }

/* Add a network socket */
    fd = PX_FD_Create_network_socket( (char *)0, netsocket, 1, &sock_info );
    if (fd < 0) {
	fprintf( stderr, "Error creating network socket\n" );
	}
    else {
	getconnection = PX_Handler_simple_create( GetConnection, sock_info );
	PX_FD_add( fd, PX_READ, getconnection, (void *)0, (void *)0, 
		  (struct timeval *)0, &ctx );
	PX_Handler_free( getconnection );
	}

/* gselect is basically an infinite loop */
    if (PX_Gselect( &ctx, (void *)0 )) {
	perror( "Gselect failed" );
	return 1;
    }
    return 0;
}

/* This is called when a listen allows an accept (either local or network) */
int GetConnection( ctx, conn, handler, in_mask )
PX_Ctx     *ctx;
PX_Conn    *conn;
PX_Handler handler;
int        in_mask;
{
    int fd;
    PX_Handler err_fd, close_fd, read_command;

    fd = PX_FD_Accept_connection( conn->fd, handler->extra_data );
    if (fd < 0) {
	fprintf( stderr, "accept errno = %d\n", errno );
	perror( "Accept failed" );
	return fd;
    }

    close_fd     = PX_Handler_simple_create( PX_Hfcn_close_fd, (void *)0 );
    err_fd       = PX_Handler_simple_create( PX_Hfcn_error_fd, (void *)0 );
    read_command = PX_Handler_simple_create( ReadCommand, 
					     InitReadCommand(ctx, fd) );
    PX_FD_add( fd,  PX_READ, read_command, err_fd, close_fd, 
	       (struct timeval *)0, ctx );

    return 0;
}

int UnlinkLocal( ctx, conn, handler, err )
PX_Ctx     *ctx;
PX_Conn    *conn;
PX_Handler handler;
int        err;
{
    unlink( LOCALPORT );
    return 0;
}
