/*
 * This is a simple command-line handler that accepts a few commands.
 * It is for example purposes only.  That is why these routines do NOT
 * use the PX_ prefix
 */
#include "pexec.h"
#include <stdio.h>
#include <memory.h>
#include <string.h>
#include "srlist.h"

extern SrList *Names;
static int debug = 1;

#define MAX_BUF 128
typedef struct {
    int  curlen;
    char buf[MAX_BUF+1];
    int  result_fd;
    FILE *resultf;
    } CommandBuf;

/*
 * This is a sample handler for accepting commands from stdin.  
 */

void *InitReadCommand( ctx, out_fd )
PX_Ctx *ctx;
int    out_fd;
{
    CommandBuf *b = (CommandBuf *)MALLOC( sizeof(CommandBuf) );
    b->curlen = 0;
    b->result_fd = out_fd;
    b->resultf = fdopen( out_fd, "a+" );
    if (!b->resultf) 
	perror( "Could not reopen fd" );
    return (void *)b;
}

/* 
 Starting at *ptr, look for a string.  Find it, null terminate it, and
 return a pointer to the first character in the string.
 Strings are space-separated.
 */
char *GetString( ptr )
char **ptr;
{
    char *p1 = *ptr;
    char *str = 0;

/* Skip spaces */
    while (*p1 && isspace(*p1)) p1++;
    str = p1;
/* Skip to space */
    while (*p1 && !isspace(*p1)) p1++;
    if (*p1) 
	*p1++ = 0;
    *ptr = p1;
    return str;
}

int PrintEntry( name, entry, cbuf )
char *name;
SrEntry *entry;
void *cbuf;
{
    CommandBuf *c = (CommandBuf *)cbuf;
    fprintf( c->resultf, "%s - %s\n", name, (char *)entry->extra_data );
}

int ReadCommand( ctx, conn, handler, in_out )
PX_Ctx     *ctx;
PX_Conn    *conn;
PX_Handler handler;
int        in_out;
{
    int        n, err;
    CommandBuf *cbuf = (CommandBuf *)handler->extra_data;
    char       *p, *p1, *cmd, *key, *name;
    SrEntry    *entry;
    int        rc;

/* Read a new command */
    if (cbuf->curlen == MAX_BUF) {
	cbuf->buf[MAX_BUF] = 0;
	fprintf( stderr, "Discarding command %s\n", cbuf->buf );
	cbuf->curlen = 0;
    }

    n = read( conn->fd, cbuf->buf+cbuf->curlen, MAX_BUF-cbuf->curlen );

    if (n <= 0) 
	return PX_Hfcn_do_eof( ctx, conn, handler, n );

/* Actually read something */
    cbuf->curlen += n;
    cbuf->buf[cbuf->curlen] = 0;

/* Process the command.  There may be several; process them all before 
   returning */
    while (1) {
	/* Is there a newline? If not, we'll just return to the select loop for the
	   rest of the data */
	p = strchr( cbuf->buf, '\n' );
	if (!p) return 0;
	*p = 0;

	if (debug) printf( "Processing command '%s'\n", cbuf->buf );
	/* 
	   Yes, there is a newline.  Process the command and update the buffer.
	   The commands are:
       
	   x          - exit
	   s          - show information on names
	   i key name - install key/name
	   d key      - delete key
	   l key      - lookup key
	   h          - help
	   */
	p1 = cbuf->buf;
	cmd = GetString( &p1 );
	switch(*cmd) {
	case 'x': 
	    /* Exit */
	    if (debug) printf( "Exit\n" );
	    PX_Ctx_close_all_conn( ctx );
	    break;
	case 'i':
	    /* Install */
	    key  = GetString( &p1 );
	    name = GetString( &p1 );
	    if (debug) printf( "Install '%s' with '%s'\n", key, name );
	    (Names->Insert)( Names, key, &entry );
	    entry->extra_data = (char *)malloc( strlen(name) + 1 );
	    strcpy( entry->extra_data, name );
	    break;
	case 'd':
	    /* Delete */
	    key  = GetString( &p1 );
	    if (debug) printf( "Delete key '%s'\n", key );
	    (Names->Delete)( Names, key );
	    break;
	case 'l':
	    /* Lookup */
	    key  = GetString( &p1 );
	    
	    if (debug) printf( "Lookup key '%s' - ", key );
	    if (!(Names->Lookup)( Names, key, &entry )) {
		if (debug) printf( "Found value '%s'\n",
				   (char *)(entry->extra_data) );
		rc = fprintf( cbuf->resultf, 
			 "%s - %s\n", key, (char *)(entry->extra_data) );
	    }
	    else {
		if (debug) printf( "Not found\n" );
		rc = fprintf( cbuf->resultf, "%s - not found\n", key );
	    }
	    if (rc <= 0) perror( "fprintf" );
	    fflush( cbuf->resultf );
	    break;
	case 's':
	    /* Show */
	    (Names->Walk)( Names,  PrintEntry, (void *)cbuf );
	    fflush( cbuf->resultf );
	    break;
	case 'h':
	    fprintf( cbuf->resultf, "Command summary:\n\
x          - exit\n\
s          - show information on names\n\
i key name - install key/name\n\
d key      - delete key\n\
l key      - lookup key\n\
h          - help\n" );
	    fflush( cbuf->resultf );
	    break;
	default:
	    break;
	}

	/* Move the remainder of buf to the beginning of the array.  Set curlen
	   accordingly */
	p1 = cbuf->buf;
	p++;
	while (*p) {
	    *p1++ = *p++;
	}
	*p1 = 0;
	cbuf->curlen = (int)(p1 - cbuf->buf);
    }

    return 0;
}

