#include "simlib.h"
#ifndef MAX_HOST_NAME
#define MAX_HOST_NAME 256
#endif
static char hostname[MAX_HOST_NAME];
static int  hostnameset = 0;

typedef struct { 
    void *sock_info;
    SIM_Ctx *ctx;
} SIM_Listen_data;

SIM_Ctx *SIM_Contact( my_name, partner_name )
char *my_name, *partner_name;
{
    int finder_fd;
    int wrt_fd, rd_fd, listen_fd;
    char contact_port[512];
    char host_and_port[512];
    SIM_Ctx *ctx;
    void       *sock_info;
    SIM_Listen_data *ldata;
    char *p;
    int  portnum, is_ready;
    PX_Handler getconnection;

    ctx = (SIM_Ctx *)malloc( sizeof(SIM_Ctx) );
    if (!ctx) return 0;
    ctx->read_callback = 0;
    ctx->rd_ready      = 0;
    ctx->wt_ready      = 0;

    ctx->pctx = (PX_Ctx *)malloc( sizeof(PX_Ctx) );
    if (!ctx->pctx) return 0;
    
    PX_Ctx_init( ctx->pctx );
    /* PX_Monitor_setup( ctx->pctx ); */

    ldata = (SIM_Listen_data *)malloc( sizeof(SIM_Listen_data) );
    if (!ldata) return 0;
    ldata->ctx = ctx;

    /* 
     * First step:  Create a listen port and send that port number
     * to the finder 
     */

    /* Create a network fd */
    portnum = 2157;
    do {
	listen_fd = PX_FD_Create_network_socket( (char *)0, portnum, 1, 
					     &ldata->sock_info );
	if (listen_fd < 0) {
	    fprintf( stderr, "Could not use port %d\n", portnum );
	    portnum++;
	}
    } while (listen_fd < 0);

    getconnection = PX_Handler_simple_create( SIM_GetConnection, ldata );
    PX_FD_add( listen_fd, PX_READ, getconnection, (void *)0, (void *)0, 
	       (struct timeval *)0, ctx->pctx );

    finder_fd = FD_Connect( (char *)0, 0 );
    if (!finder_fd) {
	printf( "Could not contact finder!\n" );
	exit( 1 );
    }
    if (!hostnameset) {
	gethostname( hostname, MAX_HOST_NAME );
	hostnameset = 1;
    }
    sprintf( host_and_port, "%s:%d", hostname, portnum );
    FD_Install( ctx->pctx, finder_fd, my_name, host_and_port, 0 );

    /*
     * Second step: find the host/port of our partner 
     */
    while (!FD_Lookup( ctx->pctx, finder_fd, 
		       partner_name, contact_port, 512 )) {
	sleep( 1 );
    }
    

    /*
     * Third step: Contact the partner 
     */
    /* Decode the server hostname and portnum from the contact port string */
    p = contact_port;
    while (*p && *p != ':') p++;
    if (*p) *p = 0;
    else {
	printf( "Invalid contact port '%s'\n", contact_port );
	return 0;
    }
    p++;
    portnum = atoi( p );
    /* This is nonblocking because we may be establishing the connection 
     * on the other side as well.  Alternately, the data in the finder
     * may be out-of-date (in which case we should really do another 
     * lookup).
     */
    do {
	wrt_fd = PX_FD_Create_network_conn( contact_port, portnum, 
					    0, &is_ready );
	sleep( 1 );
    } while (wrt_fd < 0);
    ctx->wrt_fd = wrt_fd;
    if (!is_ready) {
	/* Need to wait for it.  This just means waiting for
	   select to allow writing on wrt_fd */
	/* We could even ignore waiting on it and just trap write problems */
	;
    }

    /* Before we return, we should exchange a handshake to make sure 
       all is well */
    {
	PX_Handler write_sock;
	write_sock = PX_Handler_simple_create( SIM_CanWrite, ctx );
    PX_FD_add( wrt_fd, PX_WRITE, write_sock, 0, 0, (struct timeval *)0, 
	       ctx->pctx );
    }
    /* This gselect continues untile the connection handlers turn it off */
    PX_Gselect( ctx->pctx, (void *)0 );
    /* The gselect is terminated with a _stop, we re-enable the PEXEC 
       context for the next time */
    PX_Ctx_go( ctx->pctx );

    /* 
     * Delete my name before retunring (disallow subsequent connections)
     */
    FD_Delete( ctx->pctx, finder_fd, my_name );
    FD_Close( ctx->pctx, finder_fd );
    /* 
     * Close the listen port
     */
    PX_FD_free( ctx->pctx, listen_fd );
    close( listen_fd );
    return ctx;
}

void SIM_Close( simctx )
SIM_Ctx *simctx;
{
    free( simctx );
}

void SIM_ReadCallback( simctx, read_fcn )
SIM_Ctx *simctx;
void    (*read_fcn) ANSI_ARGS(( SIM_Ctx * ));
{
    PX_FD_set_active( simctx->pctx, simctx->rd_fd, PX_READ, 1 );
    simctx->read_callback = read_fcn;
}

/*
 * Data transfer routines
 */
void SIM_SendFloats( simctx, d, n )
SIM_Ctx *simctx;
float   *d;
int     n;
{
    /* printf( "Sending %d floats\n", n ); */
    if (FD_Write( simctx->wrt_fd, d, n * sizeof(float) ) != 
	n * sizeof(float)) {
	perror( "Writing" );
    }
	
}
void SIM_ReadFloats( simctx, d, n )
SIM_Ctx *simctx;
float   *d;
int     n;
{
    /* printf( "Reading %d floats\n", n ); */
    if (FD_Read( simctx->rd_fd, d, n * sizeof(float) ) != n * sizeof(float)) {
	perror( "Reading" );
    }
}
void SIM_SendInts( simctx, d, n )
SIM_Ctx *simctx;
int     *d;
int     n;
{
    /* printf( "Sending %d ints\n", n ); */
    FD_Write( simctx->wrt_fd, d, n * sizeof(int) );
}
void SIM_ReadInts( simctx, d, n )
SIM_Ctx *simctx;
int     *d;
int     n;
{
    /* printf( "Reading %d ints\n", n ); */
    FD_Read( simctx->rd_fd, d, n * sizeof(int) );
}

void SIM_Update( simctx )
SIM_Ctx *simctx;
{
    struct timeval poll;

    /* Eventually, would flush output */
    if (simctx->read_callback) {
	poll.tv_sec  = 0;
	poll.tv_usec = 0;
	PX_Gselect( simctx->pctx, &poll );
    }
}

/* This is called when a listen allows an accept (either local or network) */
int SIM_GetConnection( ctx, conn, handler, in_mask )
PX_Ctx     *ctx;
PX_Conn    *conn;
PX_Handler handler;
int        in_mask;
{
    int fd;
    PX_Handler err_fd, close_fd, read_command;
    SIM_Listen_data *ldata = (SIM_Listen_data *)handler->extra_data;

    fd = PX_FD_Accept_connection( conn->fd, ldata->sock_info );
    if (fd < 0) {
	perror( "Accept failed" );
	return fd;
    }

    close_fd     = PX_Handler_simple_create( PX_Hfcn_close_fd, (void *)0 );
    err_fd       = PX_Handler_simple_create( PX_Hfcn_error_fd, (void *)0 );
    read_command = PX_Handler_simple_create( SIM_ReadData, ldata->ctx );
    PX_FD_add( fd,  PX_READ, read_command, err_fd, close_fd, 
	       (struct timeval *)0, ctx );
    ldata->ctx->rd_fd = fd;
    /* Turn it off until the read callback is enabled */
    PX_FD_set_active( ctx, fd, PX_READ, 0 );
    
    ldata->ctx->rd_ready = 1;
    if (ldata->ctx->wt_ready) {
	PX_Ctx_stop( ldata->ctx->pctx );
	PX_FD_set_active( ldata->ctx->pctx, conn->fd, PX_READ, 0 );
    }
    
    /* printf( "Read fd is %d\n", fd ); */
    return 0;
}

int SIM_ReadData( ctx, conn, handler, in_out )
PX_Ctx     *ctx;
PX_Conn    *conn;
PX_Handler handler;
int        in_out;
{
    SIM_Ctx *simctx = handler->extra_data;

    /* Call the read callback */
    /* printf( "Called the \"readdata\" callback for fd %d\n", conn->fd ); */
    if (simctx->read_callback) 
	(simctx->read_callback)( simctx );
    return 0;
}

/* This is called when the connection is established */
int SIM_CanWrite( ctx, conn, handler, in_mask )
PX_Ctx     *ctx;
PX_Conn    *conn;
PX_Handler handler;
int        in_mask;
{
    int fd;
    PX_Handler err_fd, close_fd, read_command;
    SIM_Ctx *simctx = (SIM_Ctx *)handler->extra_data;

    simctx->wt_ready = 1;
    if (simctx->rd_ready) {
	PX_Ctx_stop( simctx->pctx );
	PX_FD_set_active( simctx->pctx, conn->fd, PX_WRITE, 0 );
    }
    return 0;
}
