/*
 * This file contains the main program for "RemoteRUN" (rrun)
 *
 * This is the approach:
 *    Contact the rrund at a known port (from commandline or env variable)
 *    and send user name and AUTH value (see below)
 *    Send run command
 *    Accept "forwarded" data on the socket; direct to stderr or stdout
 *    as required
 *    Forward stdin on the socket
 *    Terminate on error or when a status return comes from the rrund.
 *    Use the program's status return as the return value.
 *
 * Details:
 * The AUTH value is read from the user's .rrund_rc file, which must be
 * user-readable only.  The rrund was started with this same value.  This
 * approach assumes that the network traffic itself is secure, which is often
 * not the case.
 *
 * The use of a single socket both reduces the consumption of this resource
 * and provide a way to automatically reduce the data flow when either the
 * remote program or the local data sinks fall behind.
 */

/* This is used to store the remote program status in.  */
int prog_status = 0;

int main( argc, argv )
int argc;
char **argv;
{
    char *remote_host, **rargv;
    int  remote_port, rargc;
    PX_Ctx pctx;

    PX_Init( &pctx );
    
    /* Find the destination processor and form the run command */
    Rruns_Get_args( argc, argv, &remote_host, &remote_port, &rargc, &rargv );
    
    /* Connect to the remote network (with timeout) */
    if (Rrun_Create_network_conn( remote_host, remote_port, &fd )) return 1;

    /* Send the run command (complete with any authorization) */
    Rrun_Send_run_cmd( &pctx, fd, rargc, rargv );
        
    /* Gselect loop until done */
    if (PX_Gselect( &pctx, (void *)0 )) {
    	perror( "PX_Gselect failed" );
    	return 1;
        }

    /* Exit with REMOTE PROGRAMS status */
    return prog_status;
}

/* Send the user/AUTH value */
/* Should this be in a separate file so that the READ AUTH routine can know
   how it was sent? */
int Rrun_Send_Auth( pctx, fd )
PX_Ctx *pctx;
int    fd;
{
    FILE *fauth;
    char auth_buf[256];
    
    fauth = fopen( auth_path, "r" );
    if (!fauth)
        return Rrun_error( "Could not open AUTH file" );
    /* Check user-read only */
    fgets( fauth, auth_buf, 256 );

    /* Send message with logname, auth_buf */

    return 0;
}

/* Should this be in a separate file so that the READ side will agree on the
   form? */
int Rrun_send_run( pctx, fd, rargc, rargv )
PX_Ctx *pctx;
int    fd, rargc;
char   **rargv;
{
    /* Format run command.  So that arguments may contain embedded spaces,
       we send the run command with null separated args.
     */
    /* Get the total length */
    arglen = 0;
    for (i=0; i<rargc; i++) {
    	arglen += strlen( rargv[i] ) + 1;
        }
    /* We could send each arg separately, but we'll send it as a single
       lump */
    argbuf = (char *)MALLOC( arglen );
    if (!argbuf)
        return Rrun_error( "Could not allocate memory for command" );
    /* Copy into argbuf */
    p = argbuf;
    for (i=0; i<rargc; i++) {
    	strcpy( p, rargv[i] );
    	p += strlen(rargv[i]) + 1;
        }
    /* Send with non-blocking write (and free buffer when done) */
}

/*
 * This allows
 * rrund [ -port 1234 ] hostname program program-args
 * We should also use getenv(RRUND_PORT) as an option; should there
 * be a database of ports for each given host?
 */
int Rruns_Get_args( argc, argv, remote_host, remote_port, rargc, rargv )
int argc, *remote_port, *rargc;
char **argv, **remote_host, ***rargv;
{
    int i;
*remote_host = 0;
*remote_port = 1234;   /* Default remote port? */
*rargc = 0;
*rargv = 0;

/* Eventually, we'd look for args first */
/*    for (i=1; i<argc; i++) {
  	}
  	*/
    /* Skip command */
    argc--; argv++;
    while (argc && && argv[0] && argv[0][0] == '-') {
    	if (strcmp( argv[0], "-port" ) == 0) {
    	    argc--;
    	    argv++;
    	    if (argc)
    	        *remote_port = atoi( argv[0] );
    	    }
    	else {
    	    return Rrun_error( "Unrecognized option" );
    	argc--;
    	argv++;
        }
    if (argc < 2)
        return Rrun_error( "At least two arguments required" );
    *remote_host = argv[0];
    *rargc = argc - 1;
    *rargv = argv + 1;  	
    return 0;
}

int Rrun_Create_network_conn( remote_host, remote_port, nfd )
char *remote_host;
int  remote_port, *nfd;
{
   int fd, is_ready;

   fd = PX_FD_Create_network_conn( remote_host, remote_port, 0, &is_ready );
   if (!fd) {
   	return Rrun_error( "Unable to connect to host" );
   *nfd = fd;
   if (!is_ready) {
   	/* Select for RRUN_INIT_TIMEOUT for Write on fd */
       /* Use Gselect or just write a select loop? */
    	conn_handler = PX_Create_simple_handler( ... );
    	PX_FD_Add( fd, PX_WRITE, conn_handler, err_handler, eof_handler, ...);
        }
       }
   return 0;
}

int Rrun_Send_run_cmd( pctx, fd, rargc, rargv )
PX_Ctx  *pctx;
int fd, rargc;
char    **rargv;
{
    int rc;

    /* Send AUTH */
    if ( (rc = Rrun_Send_Auth( pctx, fd ))) return rc;
	
    /* Send Run command */
    if ( (rc = Rrun_send_run( pctx, fd, rargc, rargv )) ) return rc;

    /* Install handler (receive/send):
	STDIN forward to fd
        read FROM fd and, depending on type,
	write to stderr
	write to stdout
	set prog_status and stop select
     */
    
    return 0;
}
