#ifndef _MSGPASS
#define _MSGPASS
#include "pexec.h"
/*
 *  Definitions for message-passing layer (external objects)
 */

#define MSG_MAX_TAG   128
#define MSG_MAX_PROCS 128
#define MSG_MAX_GROUPS 64
/* 
 * Forward definitions
 */
typedef struct _MSG_Group *MSG_Group;
/* 
 * This is the basic state.  It is like "MPI_COMM_WORLD" in MPI.
 */
typedef struct {
    PX_Ctx    *ctx;                    /* PEXEC context */
    int       dbg_flags;               /* Debugging flags */
    int       n_conn;                  /* Number of connections */
    struct _MSG_FD *(fdinfo[MSG_MAX_PROCS]); /* Actual connections */
    MSG_Group groups[MSG_MAX_GROUPS];  /* Known groups */
    char      *mastername;             /* Name and port for connecting 
					  to master */
    int       masterport;
    } MSG_State;

struct _MSG_Group {
    int n_members;
    int lrank;
    int group_id;
    int l_to_g[MSG_MAX_PROCS];
    int (*(fcn[MSG_MAX_TAG])) ANSI_ARGS(( MSG_State, MSG_Group, void * ));
    void (*(extra_data[MSG_MAX_TAG]));
    };

/* All messages are sent with fixed size, network byte order date */
#define MSG_DATA 0x1
typedef struct {
    unsigned int len:32;
    unsigned int group_id:16;
    unsigned int msg_type:8;
    unsigned int tag:8;
    unsigned int source:32;
    int dest:32;
    } MSG_Header;


/*
 * FDin and FDout describe the state of an operation on a socket (FD).
 * Note that these are structured to allow partial reads/writes of data.
 */
typedef struct {
    int        busy;
    int        fd;
    MSG_Header header;
    int        n_read;
    MSG_State  *mctx;
    int        (*fcn)ANSI_ARGS(());
    void       *extra_data;
    } MSG_FDin;

typedef struct {
    int        busy;
    int        fd;
    MSG_Header header;
    int        n_written;
    MSG_State  *mctx;
    char       *buf;
    int        n_total;
    int        n_sent;
    } MSG_FDout;

/* Since sockets are bi-directional, we put BOTH in and out FD info together.
 * In addition, we store here some information on who started the connection,
 * and a flag that is used to control the initial handshake used to 
 * setup the connection (rather than having yet another structure for
 * this purpose).
 */
typedef struct _MSG_FD {
    MSG_FDin  fdin;
    MSG_FDout fdout;
    int       am_master;    
    int       state;
    } MSG_FD;

/* Possible return states (0 is always ok) */
#define MSG_ERR_BUSY 1

/* msginit.c */
int MSG_Master_setup     ANSI_ARGS(( PX_Ctx *, int *, MSG_State ** ));
int MSG_Master_connect   ANSI_ARGS(( PX_Ctx *, char *, int, MSG_State ** ));
void MSG_Dbg_state       ANSI_ARGS(( MSG_State * ));

/* msgpass.c */
int MSG_Send             ANSI_ARGS(( MSG_State *, void *, int, int, MSG_Group, 
				     int ));
int MSG_Hfcn_incoming    ANSI_ARGS(( PX_Ctx *, PX_Conn *, PX_Handler, int ));
int MSG_Hfcn_outgoing    ANSI_ARGS(( PX_Ctx *, PX_Conn *, PX_Handler, int ));

#endif





